﻿
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace XELF {

	static class Eigensystem {
		static public void Eigenvectors(
			this double[,] _matrix,
			double[] values, double[,] vectors) {
			double[,] matrix = (double[,])_matrix.Clone();

			for (int i = 0; i < 2; i++) {
				for (int j = 0; j < 2; j++) {
					vectors[i, j] = i == j ? 1 : 0;
				}
			}
			for (int l = 0; l < 2 * 2; l++) {
				double maximumElement = 0;
				int p = 0;
				int q = 0;
				for (int i = 0; i < 2; i++) {
					for (int j = 0; j < i; j++) {
						double element = Math.Abs(matrix[i, j]);
						if (element > maximumElement) {
							maximumElement = element;
							p = i;
							q = j;
						}
					}
				}

				if (maximumElement < 1e-10) break;

				double pp = matrix[p, p];
				double qq = matrix[q, q];
				double pq = matrix[p, q];
				double r = Math.Sqrt((pp - qq) * (pp - qq) + 4 * pq * pq);
				double t = Math.Atan2(-2 * pq, pp - qq) / 2;
				double cost = Math.Cos(t);
				double sint = Math.Sin(t);

				matrix[p, q] = matrix[q, p] = 0;
				matrix[p, p] = (pp + qq + r) / 2;
				matrix[q, q] = (pp + qq - r) / 2;

				for (int k = 0; k < 2; k++) {
					if (k != p && k != q) {
						double pk = matrix[p, k] * cost - matrix[q, k] * sint;
						double qk = matrix[p, k] * sint + matrix[q, k] * cost;
						matrix[p, k] = matrix[k, p] = pk;
						matrix[q, k] = matrix[k, q] = qk;
					}
				}

				for (int k = 0; k < 2; k++) {
					double pk = vectors[p, k] * cost - vectors[q, k] * sint;
					double qk = vectors[p, k] * sint + vectors[q, k] * cost;
					vectors[p, k] = pk;
					vectors[q, k] = qk;
				}
			}
			for (int i = 0; i < 2; i++) values[i] = matrix[i, i];
		}
	}

	static class Helper {
		/// <summary>
		/// 分散行列を求める。
		/// </summary>
		/// <param name="points">点群</param>
		/// <returns>分散行列</returns>
		public static double[,] VarianceMatrix(this List<Vector2> points, out Vector2 centroid) {
			centroid = points.Centroid();

			int count = points.Count;
			Vector2 r;
			double[,] matrix = new double[2, 2];

			for (int i = 0; i < count; i++) {
				Vector2 point = points[i];
				Vector2.Subtract(ref point, ref centroid, out r);
				matrix[0, 0] += r.X * r.X;
				matrix[0, 1] += r.X * r.Y;
				matrix[1, 0] += r.Y * r.X;
				matrix[1, 1] += r.Y * r.Y;
			}

			float scale = 1.0f / count;
			matrix[0, 0] *= scale;
			matrix[0, 1] *= scale;
			matrix[1, 0] *= scale;
			matrix[1, 1] *= scale;
			return matrix;
		}

		/// <summary>
		/// OBB（Oriented Bounding Box）を表す行列を計算する。
		/// 各軸の成分が-1または+1の組み合わせの頂点のからできた正方形(2Dの場合)または立方体(3Dの場合)を、
		/// この行列で変換したときに境界箱になる。
		/// </summary>
		/// <param name="points">点群</param>
		/// <returns>OBB（Oriented Bounding Box）を表す行列</returns>
		public static Matrix OBB(this List<Vector2> points) {
			Vector2 centroid;
			var matrix = points.VarianceMatrix(out centroid);
			double[] values = new double[2];
			double[,] vectors = new double[2, 2];
			matrix.Eigenvectors(values, vectors);

			var e0 = new Vector2((float)vectors[0, 1], -(float)vectors[0, 0]);
			var e1 = new Vector2((float)vectors[1, 1], -(float)vectors[1, 0]);
			var rotation = new Matrix(
				e0.X, e0.Y, 0, 0,
				e1.X, e1.Y, 0, 0,
				0, 0, 1, 0,
				0, 0, 0, 1);

			var v = Vector2.Zero;
			var min = new Vector2(float.MaxValue, float.MaxValue);
			var max = new Vector2(float.MinValue, float.MinValue);
			for (int i = 0; i < points.Count; i++) {
				var point = points[i];
				Vector2.Dot(ref e0, ref point, out v.X);
				Vector2.Dot(ref e1, ref point, out v.Y);
				Vector2.Min(ref min, ref v, out min);
				Vector2.Max(ref max, ref v, out max);
			}

			Vector2 center, scale;
			Vector2.Add(ref max, ref min, out center);
			Vector2.Subtract(ref max, ref min, out scale);
			Vector2.Multiply(ref center, 0.5f, out center);
			Vector2.Multiply(ref scale, 0.5f, out scale);

			Vector2.Transform(ref center, ref rotation, out center);

			Matrix transform, translation;
			Matrix.CreateScale(scale.X, scale.Y, 1, out transform);
			Matrix.Multiply(ref transform, ref rotation, out transform);
			Matrix.CreateTranslation(center.X, center.Y, 1, out translation);
			Matrix.Multiply(ref transform, ref translation, out transform);
			return transform;
		}
	}

	/// <summary>
	/// 2Dの主成分分析(Principal Component Analysis / PCA)
	/// </summary>
	class PrincipalComponentAnalysis2 {
		double[] values = new double[2];
		double[,] vectors = new double[2, 2];
		Vector2 centroid;

		public Vector2 Principal1 {
			get { return new Vector2((float)vectors[0, 0], (float)vectors[0, 1]); }
		}
		public Vector2 Principal2 {
			get { return new Vector2((float)vectors[1, 0], (float)vectors[1, 1]); }
		}
		public double[] Eigenvectors {
			get { return values; }
		}

		public Vector2 Centroid { get { return centroid; } protected set { centroid = value; } }

		public void Compute(List<Vector2> points) {
			var matrix = points.VarianceMatrix(out centroid);
			matrix.Eigenvectors(values, vectors);
		}
	}

}
