﻿
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XELF {

	/// <summary>
	/// 点・線・楕円・矩形の描画を提供する。
	/// </summary>
	class VectorGraphics : DrawableGameComponent {
		BasicEffect effect;
		VertexDeclaration declaration;
		readonly List<VertexPositionColor> vertices = new List<VertexPositionColor>();
		SpriteBatch spriteBatch;

		public VectorGraphics(Game game)
			: base(game) {
		}
		protected override void LoadContent() {
			this.effect = new BasicEffect(GraphicsDevice, null);
			declaration = new VertexDeclaration(GraphicsDevice, VertexPositionColor.VertexElements);
			spriteBatch = new SpriteBatch(GraphicsDevice);
			base.LoadContent();
		}

		protected override void UnloadContent() {
			declaration.Dispose();
			effect.Dispose();
			base.UnloadContent();
		}

		public Camera Camera { get; set; }

		public void Begin() {
			var viewport = GraphicsDevice.Viewport;

			spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Texture, SaveStateMode.SaveState);
		}

		public void DrawClosedLine(Vector2[] points, Color color) {
			var length = points.Length;
			if (length <= 0) return;
			for (int i = 0; i < length; i++) {
				DrawLine(ref points[i], color, ref points[(i + 1) % length], color);
			}
		}
		public void DrawClosedLine(Vector3[] points, Color color) {
			var length = points.Length;
			if (length <= 0) return;
			for (int i = 0; i < length; i++) {
				DrawLine(ref points[i], color, ref points[(i + 1) % length], color);
			}
		}

		public void DrawLine(
			Vector3 startPosition, Color startColor,
			Vector3 endPosition, Color endColor) {
			DrawLine(ref startPosition, startColor, ref endPosition, endColor);
		}
		public void DrawLine(
			ref Vector3 startPosition, Color startColor,
			ref Vector3 endPosition, Color endColor) {
			mode = Mode.Line;
			vertices.Add(new VertexPositionColor(startPosition, startColor));
			vertices.Add(new VertexPositionColor(endPosition, endColor));
		}

		public void DrawLine(
			Vector2 startPosition, Color startColor,
			Vector2 endPosition, Color endColor) {
			DrawLine(ref startPosition, startColor, ref endPosition, endColor);
		}
		public void DrawLine(
			ref Vector2 startPosition, Color startColor,
			ref Vector2 endPosition, Color endColor) {
			mode = Mode.Line;

			vertices.Add(new VertexPositionColor(new Vector3(startPosition, 1), startColor));
			vertices.Add(new VertexPositionColor(new Vector3(endPosition, 1), endColor));
		}

		public void DrawPoint(
			Vector2 position, Color color) {
			mode = Mode.Point;
			vertices.Add(new VertexPositionColor(new Vector3(position, 1), color));
		}
		Mode _mode;
		Mode mode {
			get {
				return _mode;
			}
			set {
				if (_mode != value) {
					Flush();
					Begin();
					_mode = value;
				}
			}
		}

		enum Mode {
			None,
			Point,
			Line,
			Sprite,
		}

		static readonly Vector2[] ellipseVertices;
		static readonly Vector2[] rectangleVertices =
		{
			new Vector2(-1, -1),
			new Vector2(+1, -1),
			new Vector2(+1, +1),
			new Vector2(-1, +1),
		};

		static VectorGraphics() {
			ellipseVertices = new Vector2[64];
			var a0 = (Math.PI * 2) / ellipseVertices.Length;
			for (int i = 0; i < ellipseVertices.Length; i++) {
				var a = a0 * i;
				ellipseVertices[i] = new Vector2((float)Math.Cos(a), (float)Math.Sin(a));
			}
		}

		/// <summary>
		/// 矩形を変換行列で変換して描画する。元の頂点は、(-1, -1), (+1, -1), (+1, +1), (-1, -1)。
		/// </summary>
		/// <param name="transform">変換行列</param>
		/// <param name="color">円の色</param>
		public void DrawRectangle(Matrix transform, Color color) {
			Vector2 p0 = rectangleVertices[rectangleVertices.Length - 1];
			Vector2.Transform(ref p0, ref transform, out p0);
			for (int i = 0; i < rectangleVertices.Length; i++) {
				Vector2 p1 = rectangleVertices[i];
				Vector2.Transform(ref p1, ref transform, out p1);
				DrawLine(p0, color, p1, color);
				p0 = p1;
			}
		}

		/// <summary>
		/// 原点を中心とする半径1の円を変換行列で変換して描画する。
		/// </summary>
		/// <param name="transform">変換行列</param>
		/// <param name="color">円の色</param>
		public void DrawEllipse(Matrix transform, Color color) {
			Vector2 p0 = ellipseVertices[ellipseVertices.Length - 1];
			Vector2.Transform(ref p0, ref transform, out p0);
			for (int i = 0; i < ellipseVertices.Length; i++) {
				Vector2 p1 = ellipseVertices[i];
				Vector2.Transform(ref p1, ref transform, out p1);
				DrawLine(p0, color, p1, color);
				p0 = p1;
			}
		}
		public void DrawEllipse(Vector2 center, float radius, Color color) {
			Vector2 size = new Vector2(radius);
			DrawEllipse(center - size, center + size, color);
		}
		public void DrawEllipse(Vector2 center, Vector2 size, float rotation, Color color) {
			var transform = Matrix.CreateScale(size.X, size.Y, 1)
				* Matrix.CreateRotationZ(rotation) * Matrix.CreateTranslation(center.X, center.Y, 1);
			DrawEllipse(transform, color);
		}
		public void DrawEllipse(Vector2 position0, Vector2 position1, Color color) {
			Vector2 center = position0 + position1;
			Vector2 size = position1 - position0;
			Vector2.Multiply(ref center, 0.5f, out center);
			Vector2.Multiply(ref size, 0.5f, out size);
			Vector2 p0 = ellipseVertices[ellipseVertices.Length - 1] * size + center;
			for (int i = 0; i < ellipseVertices.Length; i++) {
				Vector2 p1 = ellipseVertices[i] * size + center;
				DrawLine(p0, color, p1, color);
				p0 = p1;
			}
		}

		public void End() {
			Flush();
		}
		void Flush() {
			if (vertices.Count > 0) {
				effect.World = Camera.World;
				effect.View = Camera.View;
				effect.Projection = Camera.Projection;
				effect.VertexColorEnabled = true;
				effect.LightingEnabled = false;
				effect.Begin();
				effect.CurrentTechnique.Passes[0].Begin();
				GraphicsDevice.VertexDeclaration = declaration;

				switch (mode) {
				case Mode.Point:
					GraphicsDevice.RenderState.PointSize = 5;
					GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.PointList,
						vertices.ToArray(), 0, vertices.Count);
					break;
				case Mode.Line:
					GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList,
						vertices.ToArray(), 0, vertices.Count / 2);
					break;
				case Mode.Sprite:
					break;
				}
				_mode = Mode.None;
				vertices.Clear();
				effect.CurrentTechnique.Passes[0].End();
				effect.End();
			}
			spriteBatch.End();
		}

		public void DrawSprite(Texture2D texture, Vector2 position, Rectangle? sourceRectangle, Color color) {
			spriteBatch.Draw(texture, position, sourceRectangle, color);
		}
	}

	class Camera {
		public Matrix World = Matrix.Identity;
		public Matrix View = Matrix.Identity;
		public Matrix Projection = Matrix.Identity;
	}

}
